/***
   creddy.c
   main engine for creddy
***/

#include <assert.h>
#include <errno.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "creddy.h"

#define OPT_CN              1
#define OPT_VALIDITY        2
#define OPT_CERT            3
#define OPT_ISSUER          4
#define OPT_KEY             5
#define OPT_ROLE            6
#define OPT_SUBJECT_CERT    7
#define OPT_SUBJECT_ROLE    8
#define OPT_OUT             9 // the world oughta be opt-in only :(
#define OPT_ATTRCERT        10
#define OPT_SUBJECT_ID      11
#define OPT_SHOW            12
#define OPT_OSET            13
#define OPT_SUBJECT_OSET    14 
#define OPT_SUBJECT_OBJ     15 
#define OPT_SUBJECT_LINK    16 
#define OPT_KEYSTORE        17 
#define OPT_PASSPHRASE      18 

subject_t *subjects = NULL;
int num_subjects = 0;
int subjects_size = 0;

char **subj_items = NULL;
ENUM_CRED_TYPE *subj_items_type = NULL;
int num_subj_items = 0;
int subj_items_size = 0;

void *xmalloc(size_t);

void subject(char *subject, ENUM_SUBJ_TYPE type) {
    if (num_subjects == subjects_size) {
        subjects_size *= 2;
        subjects = xrealloc(subjects, sizeof(subject_t) * subjects_size);
    }

    int i = num_subjects++;
    subjects[i].id = NULL;
    subjects[i].cert = NULL;
    subjects[i].obj = NULL;
    subjects[i].link = NULL;

    subjects[i].cred_type = e_cred_none; // default
    subjects[i].role = NULL;

    subjects[i].subj_type = type;
    switch (type) {
      case e_subj_cert:
        subjects[i].cert = subject;
        break;
      case e_subj_id:
        subjects[i].id = subject;
        break;
      case e_subj_obj:
        subjects[i].obj = subject;
        break;
    }
}

void subj_item(char *item, ENUM_CRED_TYPE type) {
    if (num_subj_items == subj_items_size) {
        subj_items_size *= 2;
        subj_items = xrealloc(subj_items, sizeof(char *) * subj_items_size);
        subj_items_type = xrealloc(subj_items_type, 
                                  sizeof(ENUM_CRED_TYPE) * subj_items_size);
    }

    int i = num_subj_items++;
    subj_items[i] = item;
    subj_items_type[i] = type;
}

void link(char *role) {
    /* use the last subject, should be cred type */
    int i = num_subjects-1;
    subjects[i].link = role;
}
void role(char *role) {
    subj_item(role, e_cred_role);
}

void oset(char *oset) {
    subj_item(oset, e_cred_oset);
}

int main(int argc, char **argv) {

    libabac_init();
    int ret;
    options_t options = { 0, };
    char *validity_str = NULL;

    subjects = (subject_t *) xmalloc(sizeof(subject_t) * 2);
    subjects_size = 2;

    subj_items = (char **) xmalloc(sizeof(char *) * 2);
    subj_items_type = (ENUM_CRED_TYPE *) xmalloc(sizeof(ENUM_CRED_TYPE) * 2);
    subj_items_size = 2;

    struct option getopts[] = {
        { "help",       0, &options.help, 1 },
        { "generate",   0, &options.mode, MODE_GENERATE },
        { "verify",     0, &options.mode, MODE_VERIFY },
        { "keyid",      0, &options.mode, MODE_KEYID },
        { "attribute",  0, &options.mode, MODE_ATTRIBUTE },
        { "roles",      0, &options.mode, MODE_ROLES },
        { "osets",      0, &options.mode, MODE_OSETS },
        { "version",    0, &options.mode, MODE_VERSION },
        { "display",    0, &options.mode, MODE_DISPLAY },
        { "keycheck",   0, &options.mode, MODE_KEYCHECK },

        { "cert",       1, 0, OPT_CERT },
        { "p",          2, 0, OPT_PASSPHRASE },

        // generate options
        { "cn",         1, 0, OPT_CN },
        { "validity",   1, 0, OPT_VALIDITY },

        // attribute options
        { "issuer",     1, 0, OPT_ISSUER },
        { "key",        1, 0, OPT_KEY },
        { "role",       1, 0, OPT_ROLE },
        { "oset",       1, 0, OPT_OSET },
        { "subject-cert", 1, 0, OPT_SUBJECT_CERT },
        { "subject-id",   1, 0, OPT_SUBJECT_ID },
        { "subject-role", 1, 0, OPT_SUBJECT_ROLE },
        { "subject-oset", 1, 0, OPT_SUBJECT_OSET },
        { "subject-obj",  1, 0, OPT_SUBJECT_OBJ },
        { "subject-link", 1, 0, OPT_SUBJECT_LINK },
        { "keystore",     1, 0, OPT_KEYSTORE },
        { "out",          1, 0, OPT_OUT },

        // verify option
        { "attrcert",   1, 0, OPT_ATTRCERT },

        // display options
        { "show",       1, 0, OPT_SHOW },

        { NULL },
    };

    options.passphrase=0;
    for ( ; ; ) {
        int c = getopt_long(argc, argv, "", getopts, NULL);
        if (c < 0)
            break;

        switch (c) {
            // set the option from the value in the getopts struct
            case 0:
                continue;

            case OPT_CERT:
                options.cert = xstrdup(optarg);
                break;

            // for keycheck and for generate and attribute
            case OPT_PASSPHRASE:
                options.passphrase=1;
                options.pfile=NULL;
                if(optarg)
                    options.pfile=xstrdup(optarg);
                break;

            // generate options
            case OPT_CN:
                options.cn = xstrdup(optarg);
                break;
            case OPT_VALIDITY: // also an attribute option
                validity_str = xstrdup(optarg);
                break;

            // attribute options
            case OPT_ISSUER:
                options.issuer = xstrdup(optarg);
                break;
            // for keycheck and for generate
            case OPT_KEY:
                options.key = xstrdup(optarg);
                break;
            case OPT_ROLE:
                options.role = xstrdup(optarg);
                options.cred_type = e_cred_role;
                break;
            case OPT_OSET:
                options.oset = xstrdup(optarg);
                options.cred_type = e_cred_oset;
                break;
            case OPT_SUBJECT_CERT:
                subject(xstrdup(optarg), e_subj_cert);
                break;
            case OPT_SUBJECT_ID:
                subject(xstrdup(optarg), e_subj_id);
                break;
            case OPT_SUBJECT_OBJ:
                subject(xstrdup(optarg), e_subj_obj);
                break;
            case OPT_SUBJECT_LINK:
                link(xstrdup(optarg));
                break;
            case OPT_SUBJECT_ROLE:
                role(xstrdup(optarg));
                break;
            case OPT_SUBJECT_OSET:
                oset(xstrdup(optarg));
                break;
            case OPT_KEYSTORE:
                options.keystore=xstrdup(optarg);
                break;
            case OPT_OUT:
                options.out = xstrdup(optarg);
                break;

            // verify options
            case OPT_ATTRCERT:
                options.attrcert = xstrdup(optarg);
                break;

            // display options
            case OPT_SHOW:
                options.show = xstrdup(optarg);
                break;

            case '?':
                break;

            default:
                printf("wat\n");
                return 45;
        }
    }

    if (options.help || optind < argc) {
        if (optind > 0 && optind < argc)
            printf("I don't understand %s, %d\n", argv[optind], optind);
        usage(&options);
    }

    // parse the validity
    if (validity_str != NULL) {
        char suffix = 'd'; // default suffix is days
        int multiplier;

        int len = strlen(validity_str);
        assert(len > 0);

        // get the suffix char if it's alphabetical
        if (isalpha(validity_str[len - 1])) {
            suffix = validity_str[len - 1];

            // truncate
            validity_str[len - 1] = '\0';
            --len;

            // make sure it's not only a suffix
            if (len == 0) {
                printf("Invalid validity\n");
                usage(&options);
            }
        }

        // convert the suffix to a multiplier
        switch(suffix) {
            case 's': multiplier =        1; break;
            case 'm': multiplier =       60; break;
            case 'h': multiplier =     3600; break;
            case 'd': multiplier =    86400; break;
            case 'y': multiplier = 31536000; break;
            default:
                printf("Invalid suffix, must be s m h d y\n");
                usage(&options);
        }

        // ascii to int
        char *end;
        options.validity = strtol(validity_str, &end, 10);
        if (errno != 0 || end - validity_str < len) {
            printf("Invalid validity\n");
            usage(&options);
        }

        if (options.validity <= 0) {
            printf("Invalid validity: must be > 0\n");
            usage(&options);
        }

        // multiply!
        options.validity *= multiplier;

        free(validity_str);
    }

    if (options.mode == MODE_ATTRIBUTE) {
        int i;
        int j;

        // have to do error checking on subjects here
        if (
                (num_subjects == 0) ||
                (num_subjects != (num_subj_items) && num_subjects != 1 
                                 && (num_subj_items) != 0)
           ) {
            printf(
                "You have %d subject%s and %d role/oset%s, which is mismatched\n",
                num_subjects, num_subjects == 1 ? "" : "s",
                (num_subj_items), (num_subj_items) == 1 ? "" : "s"
            );
            usage(&options);
        }

        for (i = 0; i < num_subj_items; ++i) {
            subjects[i].cred_type=subj_items_type[i];
            if(subj_items_type[i] == e_cred_role)
                subjects[i].role = subj_items[i];
                else subjects[i].oset=subj_items[i];
        }
        free(subj_items);

        options.subjects = subjects;
        options.num_subjects = num_subjects;
    }

    // launch the sub command
    switch (options.mode) {
        case MODE_GENERATE:
            if (options.validity == 0) options.validity = 1080 * 86400;
            generate_main(&options);
            break;

        case MODE_KEYCHECK:
            keycheck_main(&options);
            break;

        case MODE_KEYID:
            keyid_main(&options);
            break;

        case MODE_ATTRIBUTE:
            if (options.validity == 0) options.validity = 365 * 86400;
           attribute_main(&options);
            break;

        case MODE_ROLES:
            roles_main(&options);
            break;

        case MODE_OSETS:
            osets_main(&options);
            break;

        case MODE_VERIFY:
            verify_main(&options);
            break;

        case MODE_DISPLAY:
            display_main(&options);
            break;

        case MODE_VERSION:
            printf("ABAC/creddy " ABAC_VERSION "\n");
            break;

        default:
            usage(&options);
    }

    return 0;
}

void usage(options_t *opts) {
    if (opts->mode == MODE_GENERATE)
        printf(
            "Usage: creddy --generate --cn <name> [ --validity <time> ] [ --out <dir> ] [ --key <keyfile> ] [ --p <=pfile> ] \n"
            "    generated cert will be in ${name}_ID.der, \n"
            "    a generated private key in ${name}_private.pem \n"
            "       unless a keyfile is supplied, \n"
            "    if supplied keyfile is encrypted, \n"
            "       a passphrase can be ingested through a pfile, \n"
            "       or it will be prompted for interactively,  \n"
            "    files output to dir if specified\n"
            "    default validity: 1080 days\n"
            "\n"
            "    time is specified with optional suffix: s m h d y\n"
            "    defaults to days if unspecified\n"
        );

    else if (opts->mode == MODE_VERIFY)
        printf(
            "Usage: creddy --verify --cert <cert> [ --attrcert <cert> ]\n"
            "    if attrcert is provided, verify that it was issued by cert\n"
            "    if attrcert is the same as cert, verify for self-signing cert\n"
        );

    else if (opts->mode == MODE_KEYID)
        printf(
            "Usage: creddy --keyid --cert <cert>\n"
        );

    else if (opts->mode == MODE_KEYCHECK)
        printf(
            "Usage: creddy --keycheck --key <key> --p <=pfile>\n"
            "   check if the private key is accessible, \n"
            "   if passphrase is supplied, it is used to decrypt the key\n"
        );

    else if (opts->mode == MODE_ATTRIBUTE)
        printf(
            "Usage: creddy --attribute \\\n"
            "                --issuer <cert> --key <key> --p <=pfile>\\\n"
            "                [ --role <role> | --oset <oset> ] \\\n"
            "                [ --subject-cert <cert> | --subject-id <sha1> | --subject-obj <object> ] \\\n"
            "                [ --subject-link <role> | --subject-role <role> | --subject-oset <oset> ]  ... ] \\\n"
            "                [ --validity <time> ] --out <file>\n"
            "    default validity: 365 days\n"
            "    provide exactly one of --subject-cert / --subject-id / --subject-obj\n"
            "    give multiple --subject-{cert,id} / --subject-{role,obj} pairs for intersection\n"
            "\n"
            "    time is specified with optional suffix: s m h d y\n"
            "    defaults to days if unspecified\n"
        );

    else if (opts->mode == MODE_ROLES)
        printf(
            "Usage: creddy --roles --cert <cert>\n"
        );

    else if (opts->mode == MODE_OSETS)
        printf(
            "Usage: creddy --osets --cert <cert>\n"
        );

    else if (opts->mode == MODE_DISPLAY)
        printf(
            "Usage: creddy --display --show=[issuer,..,all] --cert <cert>\n"
            "   values for --show are comma-separated:\n"
            "       issuer      DN of issuer\n"
            "       subject     DN of subject\n"
            "       validity    validity period\n"
            "       roles       attribute cert roles (fails silently on ID certs)\n"
            "       osets       attribute cert osets (fails silently on ID certs)\n"
            "       all         all of the above\n"
            "   cert may be X.509 identity or attribute cert\n"
        );

    else
        printf(
            "Usage: creddy [ --<mode> ] [ --help ]\n"
            "    --generate:  generate X.509 identity cert and optional private key\n"
            "    --verify:    check validity of X.509 ID or attribute cert\n"
            "    --keyid:     get fingerprint from X.509 ID cert\n"
            "    --attribute: generate an X.509 attribute cert\n"
            "    --roles:     list roles from an X.509 attribute cert\n"
            "    --osets:     list osets from an X.509 attribute cert\n"
            "    --display:   list metadata from an X.509 identity or attribute cert\n"
            "    --keycheck:  test the private key\n"
            "    --version:   display ABAC version\n"
        );


    exit(1);
}

void *xmalloc(size_t len) {
    void *ret = malloc(len);
    if (ret == NULL)
        err(1, "couldn't malloc %d bytes\n", len);
    return ret;
}

void *xrealloc(void *ptr, size_t size) {
    void *ret = realloc(ptr, size);
    if (ret == NULL)
        err(1, "couldn't realloc %d bytes\n", size);
    return ret;
}

char *xstrdup(char *string) {
    char *dup = strdup(string);
    if (dup == NULL)
        err(1, "Can't dup %s", string);
    return dup;
}
